import datetime

from croniter import croniter
from pydantic_db_backend_common.utils import utcnow

from eventix.pydantic.schedule import Schedule


def schedule_set_next_schedule(schedule: Schedule):
    if schedule.next_schedule is None or schedule.next_schedule < utcnow():
        schedule.last_schedule = schedule.next_schedule
        schedule.next_schedule = croniter(schedule.schedule).get_next(datetime.datetime, utcnow())




from collections import OrderedDict
from pprint import pprint
from typing import Optional, List, Dict, Callable

import click
import lsjsonclasses
import pydash
from lsrestclient import LsRestClient
from pydantic import BaseModel

from eventix.functions.tools import version
from eventix.pydantic.task import TaskModel


class CLIContext(BaseModel):
    client: LsRestClient
    namespace: Optional[str]

    class Config:
        arbitrary_types_allowed = True


@click.group()
@click.version_option(version())
@click.option("-s", "--server", help="server url")
@click.option("-n", "--namespace", help="namespace")
@click.pass_context
def cli(ctx, server: str = None, namespace: str = None):
    ctx.obj = CLIContext(
        namespace=namespace,
        client=LsRestClient(base_url=server, name="eventix")
    )


@cli.group("get")
@click.pass_context
def cli_get(ctx):
    pass


@cli_get.command("tasks")
@click.option("--limit", "limit", default=10, type=int)
@click.option("--skip", "skip", default=0, type=int)
@click.option("--status", "status", default=None, type=str)
@click.pass_context
def cli_get_tasks(ctx, skip: int, limit: int, status: str | None = None):
    namespace = ctx.obj.namespace

    body = dict(
        skip=skip,
        limit=limit
    )

    body |= dict(
        namespace=namespace,
        status=status
    )
    # pprint(body)

    r = ctx.obj.client.put("/tasks/by_status", body=body)
    if r.status_code == 200:
        pv = r.json()
        max_result = pydash.default_to(pydash.get(pv, "max_results"), 0)

        start = skip + 1
        end = skip + limit if skip + limit < max_result else max_result
        pagination_result = f"{start}-{end} of {max_result}"

        tasks = [TaskModel.parse_obj(x) for x in pydash.get(pv, 'data', [])]

        if len(tasks) != 0:
            entries = [
                OrderedDict(
                    NAMESPACE=task.namespace,
                    UID=task.uid,
                    TASK=task.task,
                    STATUS=task.status,
                    IDENTIFIER=""
                ) for task in tasks
            ]
            print_table(entries, click.echo)

            click.echo(f"Results {pagination_result}")
        else:
            click.echo(f"No result")


@cli_get.command("task")
@click.option("--error-only", is_flag=True, default=False)
@click.argument("uid")
@click.pass_context
def cli_get_task(ctx, uid: str, error_only: bool):
    r = ctx.obj.client.get("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        data = r.json()
        result = pydash.get(data, "result", None)
        if isinstance(result, dict):
            error_class = pydash.get(result, "error_class", None)
        else:
            error_class = None

        if not error_only:
            if error_class is not None:
                data['result'] = pydash.pick(result, "error_class")

            print_dict_to_json(data, click.echo)
        else:
            click.echo(f"ERROR_CLASS: {error_class}")
            if error_class == "LSoftException":
                error_dict = lsjsonclasses.LSoftJSONDecoder.loads(result['error_message'])
                error = pydash.get(error_dict, "detail.ERROR", None)
                traceback_str = pydash.get(error_dict, "detail.TRACEBACK", None)
                click.echo(f"ERROR:")
                click.echo(error)
                click.echo(f"TRACEBACK:")
                click.echo(traceback_str)
            else:
                print_dict_to_json(result, click.echo)

    else:
        click.echo(f"Error: {r.content}")


@cli.group("delete")
@click.pass_context
def cli_delete(ctx):
    pass


@cli_delete.command("task")
@click.argument("uid")
@click.pass_context
def cli_get_task(ctx, uid: str):
    r = ctx.obj.client.delete("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        click.echo(f"Deleted task {uid}")
    else:
        click.echo(f"Error: {r.content}")


def print_dict_to_json(data: dict, print_func: Callable = print):
    j = lsjsonclasses.LSoftJSONEncoder.dumps(data, indent=2)
    print_func(j)


def print_table(tdata: List[Dict[str, str]], print_func: Callable = print):
    if len(tdata) == 0:
        return

    keys = list(tdata[0].keys())
    sizes = {}
    lines = []
    for key in keys:
        sizes[key] = max(pydash.concat([len(key)], [len(x[key]) for x in tdata]))

    tdata = pydash.concat([{k: k for k in keys}], tdata)
    for td in tdata:
        line = ""
        for key in keys:
            line += "{text:{width}} ".format(text=td[key], width=sizes[key])
        lines.append(line.strip())

    for line in lines:
        print_func(line)


if __name__ == '__main__':
    cli()



import logging
import traceback
from typing import Dict, Optional, Any, Type

from fastapi import Request, HTTPException
from pydantic_db_backend.backend import Backend, BackendBase
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction, RequestResponseEndpoint
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp
from webexception.webexception import WebException

log = logging.getLogger(__name__)


class BackendMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        dispatch: DispatchFunction | None = None,
        backend: Type[BackendBase] = None
    ) -> None:
        super().__init__(app, dispatch)
        self.backend = backend

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        with Backend.provider(self.backend):
            response = await call_next(request)
        return response


class ErrorMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        dispatch: DispatchFunction | None = None,
    ) -> None:
        super().__init__(app, dispatch)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        try:
            response = await call_next(request)
        except Exception as e:
            return await self.make_exception_response(e)
        return response

    async def make_exception_detail(self, e: Exception) -> Dict:
        tb = traceback.format_exc()
        ret = dict(error_class=e.__class__.__name__, error_message=str(e), error_traceback=tb)
        return ret

    async def make_exception_response(self, e: Exception, headers: Optional[Dict[str, Any]] = None) -> JSONResponse:
        status_code = None
        if isinstance(e, HTTPException):
            status_code = e.status_code
            content = dict(detail=e.detail)

        elif isinstance(e, WebException):
            status_code = e.status_code
            if status_code == 204:
                content = None
            else:
                content = e.dict()
        else:
            if status_code is None:  # no status code found yet
                status_code = 500
            content = dict(detail=await self.make_exception_detail(e))

        if status_code >= 500:
            log.exception(e, stacklevel=2)

        if isinstance(content, dict):
            response = JSONResponse(status_code=status_code, content=content, headers=headers)
        else:
            response = Response(status_code=status_code, headers=headers, content=content)
        return response



from __future__ import annotations

from typing import Any

import requests
from requests import Session, Response

import logging

log = logging.getLogger(__name__)


# class EventixClientSession(Session):
#     def __init__(self, base_url: str = None) -> None:
#         self.base_url = base_url
#         self.keep_alive = False
#         super().__init__()
#
#     def request(
#         self,
#         method: str | bytes,
#         url: str | bytes,
#         *args,
#         **kwargs
#     ) -> Response:  # pragma: no cover
#         # kwargs['headers'] = kwargs.get('headers', {}) | {"Connection": "close"}
#         return super().request(
#             method,
#             f"{self.base_url}{url}",
#             *args,
#             **kwargs
#         )

class EventixClientSession(Session):
    def __init__(self, base_url: str = None) -> None:
        self.base_url = base_url
        super().__init__()

    def request(
        self,
        method,
        url,
        *args,
        **kwargs
    ) -> Response:  # pragma: no cover
        return requests.request(
            method,
            f"{self.base_url}{url}",
            *args,
            **kwargs
        )


def get_session():
    s = EventixClientSession()
    s.headers["Connection"] = "close"
    return s


class EventixClient:
    # interface: Any | None = EventixClientSession()
    interface: Any | None = get_session()
    namespace: str | None = None

    @classmethod
    def set_base_url(cls, base_url):
        if isinstance(cls.interface, EventixClientSession):
            log.info(f"Setting EventixClient base_url: {base_url}")
            cls.interface.base_url = base_url



import logging
from typing import List, Annotated

from fastapi import APIRouter, Body, Depends

from eventix.exceptions import NoTaskFound
from eventix.functions.task import task_post, task_next_scheduled, tasks_by_status
from eventix.pydantic.pagination import dep_pagination_parameters, PaginationParametersModel, PaginationResultModel
from eventix.pydantic.task import TaskModel
from pydantic_db_backend.backend import Backend

log = logging.getLogger(__name__)

router = APIRouter(tags=["tasks"])


@router.get("/tasks/next_scheduled")
async def route_tasks_next_scheduled_get(worker_id: str, namespace: str) -> TaskModel:
    t = task_next_scheduled(worker_id, namespace)
    if t is None:
        raise NoTaskFound(namespace=namespace)
    return t


class RouterTasksByStatusResponseModel(PaginationResultModel):
    data: List[TaskModel]


@router.put("/tasks/by_status")
async def router_tasks_by_status_put(
    status: Annotated[str, Body()] = None,
    namespace: Annotated[str, Body()] = None,
    pagination: PaginationParametersModel = Depends(dep_pagination_parameters)
) -> RouterTasksByStatusResponseModel:
    data, max_results = tasks_by_status(
        status=status,
        namespace=namespace,
        pagination=pagination,
        max_results=False
    )
    return RouterTasksByStatusResponseModel(
        data=data,
        max_results=max_results,
        **pagination.dict()
    )



import importlib
import logging
import os
from uuid import uuid4

import pkg_resources
import toml


def uid() -> str:
    return str(uuid4()).replace("-", "")


_uid = uid


def version():
    v = importlib.metadata.version('eventix')
    return v


def setup_logging(without_time: bool = False, level: str = "INFO", module: bool = False, prefix: str = ""):
    time_str = '%(asctime)s ' if not without_time else ''
    module = '[%(name)s]' if module else ''
    f = f"{time_str}{module}[%(levelname)8s]{prefix} %(message)s"

    logging.basicConfig(
        level=level,
        format=f,
        handlers=[
            logging.StreamHandler()
        ]
    )



import datetime
from typing import List, Any

from pydantic import Field, BaseModel

from pydantic_db_backend_common.pydantic import BackendModel


class Schedule(BackendModel):
    schedule: str
    task: str
    name: str | None = None
    args: List[Any] | None = Field(default_factory=list)
    kwargs: dict | None = Field(default_factory=dict)
    last_schedule: datetime.datetime | None = None
    next_schedule: datetime.datetime | None = None
    priority: int | None = 0

    def __init__(__pydantic_self__, **data: Any) -> None:
        if "name" not in data:
            data["name"] = data["task"]
        super().__init__(**data)



import datetime
import functools
import logging
from inspect import signature
from typing import Callable, Dict, Any, _BaseGenericAlias

from pydantic import BaseModel
from pydantic_db_backend_common.utils import str_to_datetime_if_parseable

from eventix.contexts import namespace_context, delay_tasks_context, task_priority_context
from eventix.functions.task_scheduler import TaskScheduler
from eventix.pydantic.task import TaskModel

log = logging.getLogger(__name__)


class EventixTaskBase(object):
    pass


def task(store_result: bool = True, unique_key_generator: Callable = None, retry: bool = True, max_retries: int | None = None, error_expires: int | None = None, result_expires: int | None = 604800):
    # parameters suggested:
    #
    # store_result: bool , wether to store or not to store results
    # unique_uid: ....  having a unique id, which leads to update the task if it is rescheduled.

    # retry: bool, wether to retry failed tasks
    # max_retry: int, maximum retries default 5 ... each retry doubles the time to wait, starting with 30 sec

    is_unique = unique_key_generator is not None

    def inner(f):
        class EventixTask(EventixTaskBase):
            def __init__(self):
                self.func = f
                self.func_name = f.__name__
                self.arg_spec = list(signature(f).parameters.values())
                self.arg_spec_args = [arg.annotation for arg in self.arg_spec]
                self.arg_spec_kwargs = {arg.name: arg.annotation for arg in self.arg_spec}
                assert True

            def get_arg_from_cls(self, cls, arg) -> Any:

                #### commented code is a first start to restructure

                # if isinstance(arg, dict):
                #     if isinstance(cls, _BaseGenericAlias):
                #         cls = cls.__origin__
                #
                # match cls:
                #     case datetime.datetime:
                #         if isinstance(arg, str):
                #             arg = str_to_datetime_if_parseable(arg)
                #     case BaseModel:
                #         if isinstance(arg, dict):
                #             arg = cls.parse_obj(arg)

                match cls:
                    case datetime.datetime:
                        # next line if is for calling task with run. in this case the arg is already a datetime
                        if isinstance(arg, str):
                            arg = str_to_datetime_if_parseable(arg)
                    case _:
                        if isinstance(arg, dict):
                            if isinstance(cls, _BaseGenericAlias):
                                cls = cls.__origin__
                            if issubclass(cls, BaseModel):
                                arg = cls.parse_obj(arg)
                return arg

            def restore_original_types(self, args, kwargs) -> tuple[list[Any], Dict[str, Any]]:
                new_args = []
                for i, arg in enumerate(args):
                    cls = self.arg_spec_args[i]
                    arg = self.get_arg_from_cls(cls=cls, arg=arg)
                    new_args.append(arg)
                new_kwargs = {}
                for kw, arg in kwargs.items():
                    cls = self.arg_spec_kwargs[kw]
                    new_kwargs[kw] = self.get_arg_from_cls(cls=cls, arg=arg)
                return new_args, new_kwargs

            @functools.wraps(f)
            def delay(self, *args, _priority: int | None = None, _eta: datetime.datetime | None = None, **kwargs) -> TaskModel:
                if _priority is None:
                    with task_priority_context() as task_priority:
                        _priority = task_priority
                # priority has to be negated, needed for fixing ascending sort order
                tm = self.make_task_model(args, kwargs, priority=_priority * -1, eta=_eta)
                with delay_tasks_context() as delay:
                    if delay:
                        tm = TaskScheduler.schedule(tm)
                    else:
                        self.run(*args, **kwargs)

                # IDEA: It could be possible to run worker execute with generated task during tests
                return tm

            @functools.wraps(f)
            def run(self, *args, **kwargs):
                args, kwargs = self.restore_original_types(args, kwargs)
                return self.func(*args, **kwargs)

            def make_task_model(self, args, kwargs, priority, eta) -> TaskModel:
                with namespace_context() as namespace:
                    params = dict(
                        task=self.func_name,
                        args=args,
                        kwargs=kwargs,
                        priority=priority,
                        store_result=store_result,
                        retry=retry,
                        max_retries=max_retries,
                        error_expires=error_expires,
                        result_expires=result_expires,
                        namespace=kwargs['namespace'] if "namespace" in kwargs else namespace
                    )
                    if eta is not None:
                        if eta.tzinfo is None:
                            eta.replace(tzinfo=datetime.timezone.utc)
                        params["eta"] = eta

                    if is_unique:
                        unique_key = unique_key_generator(*args, **kwargs)
                        params |= dict(unique_key=unique_key)

                return TaskModel.parse_obj(params)

        return EventixTask()

    return inner
